"""
x9k3/scte35.py Home of the SCTE35 class used by x9k3.

"""

import datetime


class SCTE35:
    """
    A SCTE35 instance is used to hold
    SCTE35 cue data by X9K3.
    """

    def __init__(self):
        self.cue = None
        self.cue_state = None
        self.cue_time = None
        self.tag_method = self.x_cue
        self.break_timer = None
        self.break_duration = None
        self.event_id = 1
        self.seg_type = None

    def mk_cue_tag(self):
        """
        mk_cue_tag routes hls tag creation
        """
        tag = False
        if self.cue:
            tag = self.tag_method()
        return tag

    def chk_cue_state(self):
        """
        chk_cue_state changes self.cue_state
        """
        if self.cue_state == "OUT":
            self.cue_state = "CONT"
        if self.cue_state == "IN":
            self.cue_time = None
            self.cue = None
            self.cue_state = None
            self.break_timer = None

    def mk_cue_state(self):
        """
        mk_cue_state checks if the cue
        is a CUE-OUT or a CUE-IN and
        sets cue_state.
        """
        if self.is_cue_out(self.cue):
            self.cue_state = "OUT"
            self.break_timer = 0.0
            if self.cue_time and self.break_duration:
                self.cue_time += self.break_duration
        if self.is_cue_in(self.cue):
            # the next line causes splice immediate break returns to fail.
            # if self.break_timer >= self.break_duration:
            self.cue_state = "IN"

    def x_cue(self):
        """
        #EXT-X-CUE-( OUT | IN | CONT )
        """
        if self.cue_state == "OUT":
            return f"#EXT-X-CUE-OUT:{self.break_duration}"
        if self.cue_state == "IN":
            return "#EXT-X-CUE-IN"
        if self.cue_state == "CONT":
            return f"#EXT-X-CUE-OUT-CONT:{self.break_timer:.6f}/{self.break_duration}"
        return False

    def x_splicepoint(self):
        """
        #EXT-X-SPLICEPOINT-SCTE35
        """
        base = f"#EXT-X-SPLICEPOINT-SCTE35:{self.cue.encode()}"
        if self.cue_state == "OUT":
            return f"{base}"
        if self.cue_state == "IN":
            return f"{base}"
        return False

    def x_scte35(self):
        """
        #EXT-X-SCTE35
        """
        base = f'#EXT-X-SCTE35:CUE="{self.cue.encode()}" '
        if self.cue_state == "OUT":
            return f"{base},CUE-OUT=YES "
        if self.cue_state == "IN":
            return f"{base},CUE-IN=YES "
        if self.cue_state == "CONT":
            return f"{base},CUE-OUT=CONT"
        return False

    def x_daterange(self):
        """
        #EXT-X-DATERANGE
        """
        tag = ""
        fbase = f'#EXT-X-DATERANGE:ID="{self.event_id}"'
        iso8601 = f"{datetime.datetime.utcnow().isoformat()}Z"
        fdur = ""
        if self.break_duration:
            fdur = f",PLANNED-DURATION={self.break_duration}"
        if self.cue_state == "OUT":
            fstart = f',START-DATE="{iso8601}"'
            tag = f"{fbase}{fstart}{fdur},SCTE35-OUT={self.cue.encode_as_hex()}"
            return tag
        if self.cue_state == "IN":
            fstop = f',END-DATE="{iso8601}"'
            tag = f"{fbase}{fstop},SCTE35-IN={self.cue.encode_as_hex()}"
            self.event_id += 1
            return tag
        return tag

    def _splice_insert_cue_out(self, cue):
        cmd = cue.command
        if cmd.out_of_network_indicator:
            if cmd.break_duration:
                self.break_duration = cmd.break_duration
            self.cue_state = "OUT"
            return True
        return False

    def _time_signal_cue_out(self, cue):
        seg_starts = [0x22, 0x30, 0x32, 0x34, 0x36, 0x44, 0x46]
        for dsptr in cue.descriptors:
            if dsptr.tag != 2:
                return False
            if dsptr.segmentation_type_id in seg_starts:
                self.seg_type = dsptr.segmentation_type_id + 1
                if dsptr.segmentation_duration:
                    self.break_duration = dsptr.segmentation_duration
                    self.cue_state = "OUT"
                    return True
        return False

    def is_cue_out(self, cue):
        """
        is_cue_out checks a Cue instance
        Returns True for a cue_out event.
        """
        if cue is None:
            return False
        if self.cue_state not in ["IN", None]:
            return False
        cmd = cue.command
        if cmd.command_type == 5:
            return self._splice_insert_cue_out(cue)
        if cmd.command_type == 6:
            return self._time_signal_cue_out(cue)

        return False

    def is_cue_in(self, cue):
        """
        is_cue_in checks a Cue instance
        Returns True for a cue_in event.
        """
        if cue is None:
            return False
        if self.cue_state not in ["OUT", "CONT"]:
            return False
        cmd = cue.command
        if cmd.command_type == 5:
            if not cmd.out_of_network_indicator:
                return True
        if cmd.command_type == 6:
            for dsptr in cue.descriptors:
                if dsptr.tag == 2:
                    if dsptr.segmentation_type_id == self.seg_type:
                        self.seg_type = None
                        self.cue_state = "IN"
                        return True
        return False



"""
x9k3/sliding.py
"""

from collections import deque
from os import unlink
from pathlib import Path
from threefive import blue


class SlidingWindow:
    """
    The SlidingWindow class
    """

    def __init__(self, size=50000):
        self.size = size
        self.panes = deque()
        self.delete = False

    def popleft_pane(self):
        """
        popleft_pane removes the first item in self.panes
        """
        popped = self.panes.popleft()
        if self.delete:
            Path(popped.name).touch()
            unlink(popped.name)
            blue(f"deleted {popped.name}")

    def push_pane(self, a_pane):
        """
        push appends a_pane to self.panes
        """
        self.panes.append(a_pane)

    def all_panes(self):
        """
        all_panes returns the current window panes joined.
        """
        return "".join([a_pane.get() for a_pane in self.panes])

    def slide_panes(self, a_pane=None):
        """
        slide calls self.push_pane with a_pane and then
        calls self.popleft_pane to trim self.panes as needed.
        """
        # blue(f"slide pane. \n Panes:{len(self.panes)}")
        if a_pane:
            self.push_pane(a_pane)
        if len(self.panes) > self.size:
            self.popleft_pane()



"""
x9k3/argue.py
"""

import argparse
import sys


def argue():
    """
    argue parse command line args
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        default=sys.stdin.buffer,
        help=""" The Input video  can be mpegts or m3u8 with mpegts segments,
                        or a playlist with mpegts files and/or mpegts m3u8 files.
                    The input can be a local video, http(s), udp, multicast or stdin.
            """,
    )
    parser.add_argument(
        "-b",
        "--byterange",
        action="store_const",
        default=False,
        const=True,
        help="Flag for byterange hls [default:False]",
    )
    parser.add_argument(
        "-c",
        "--continue_m3u8",
        action="store_const",
        default=False,
        const=True,
        help="Resume writing index.m3u8 [default:False]",
    )
    parser.add_argument(
        "-d",
        "--delete",
        action="store_const",
        default=False,
        const=True,
        help="delete segments (enables --live) [default:False]",
    )
    parser.add_argument(
        "-l",
        "--live",
        action="store_const",
        default=False,
        const=True,
        help="Flag for a live event (enables sliding window m3u8) [default:False]",
    )
    parser.add_argument(
        "-n",
        "--no_discontinuity",
        action="store_const",
        default=False,
        const=True,
        help="Flag to disable adding #EXT-X-DISCONTINUITY tags at splice points [default:False]",
    )
    parser.add_argument(
        "-no_adrian_is_cool_tags_at_splice_points_because_I_suck",
        action="store_const",
        default=False,
        const=True,
        help="Flag to disable adding #EXT-X-ADRIAN-IS-COOL tags at splice points [default:False]",
    )

    parser.add_argument(
        "-N",
        "--no-throttle",
        "--no_throttle",
        action="store_const",
        default=False,
        const=True,
        help=" Flag to disable live throttling [default:False]",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        default=".",
        help="Directory for segments and index.m3u8(created if needed) [default:'.']",
    )
    parser.add_argument(
        "-e",
        "--exclude_mpegts",
        action="store_const",
        default=False,
        const=True,
        help="Flag to exclude parsing SCTE-35 from MPEGTS. [default:False]",
    )
    parser.add_argument(
        "-p",
        "--program_date_time",
        action="store_const",
        default=False,
        const=True,
        help="Flag to add Program Date Time tags to index.m3u8. [default:False]",
    )
    parser.add_argument(
        "-r",
        "--replay",
        action="store_const",
        default=False,
        const=True,
        help="Flag for replay aka looping (enables --live,--delete) [default:False]",
    )
    parser.add_argument(
        "-s",
        "--sidecar_file",
        default="sidecar.txt",
        help="""Sidecar file of SCTE-35 (pts,cue) pairs. [default:sidecar.txt]""",
    )
    parser.add_argument(
        "-S",
        "--shulga",
        action="store_const",
        default=False,
        const=True,
        help="Flag to enable Shulga iframe detection mode [default:False]",
    )
    parser.add_argument(
        "-t",
        "--time",
        default=2,
        type=float,
        help="Segment time in seconds [default:2]",
    )
    parser.add_argument(
        "-T",
        "--hls_tag",
        default="x_cue",
        help="x_scte35, x_cue, x_daterange, or x_splicepoint [default:x_cue]",
    )

    parser.add_argument(
        "-w",
        "--window_size",
        default=5,
        type=int,
        help="sliding window size (enables --live) [default:5 max:10]",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="store_const",
        default=False,
        const=True,
        help="Show version",
    )
    return parser.parse_args()



"""
x9k3/__init__.py
"""

from .abr import ABR
from .x9k3 import X9K3, SegmentData, Timer, version
from .x9mp import cli
from .argue import argue
from .scte35 import SCTE35
from .sliding import SlidingWindow



#!/usr/bin/env python3

"""
x9k3/x9k3.py
"""
import datetime
import io
import os
import sys
import time
from collections import deque
from operator import itemgetter
from pathlib import Path
from threefive import blue, pif, print2, red, reader
from threefive import Cue, ERR, IFramer, Segment
import threefive.stream as strm
from m3ufu import M3uFu
from .argue import argue
from .scte35 import SCTE35
from .sliding import SlidingWindow
from threefive import blue, red

MAJOR = "1"
MINOR = "0"
MAINTAINENCE = "13"


def version():
    """
    version prints x9k3's version as a string
    Odd number versions are releases.
    """
    return f"{MAJOR}.{MINOR}.{MAINTAINENCE}"


class X9K3(strm.Stream):
    """
    X9K3 class
    """

    def __init__(self, tsdata=None, show_null=False):
        super().__init__(tsdata, show_null)
        self._tsdata = tsdata
        self.in_stream = tsdata
        self.active_segment = io.BytesIO()
        self.iframer = IFramer(shush=True)
        self.scte35 = SCTE35()
        self.sidecar = deque()
        self.timer = Timer()
        self.m3u8 = "index.m3u8"
        self.window = SlidingWindow()
        self.segnum = None
        self.args = argue()
        self.started = None
        self.next_start = None
        self.media_seq = 0
        self.discontinuity_sequence = 0
        self.first_segment = True
        self.first_on_page = None
        self.now = None
        self.last_sidelines = ""
        self.started_byte = 0
        self.now_byte = 0
        self.pat_pkt = None
        self.pmt_pkt = None
        self.skipped_segment = False

    def _args_version(self):
        if self.args.version:
            print2(version())
            sys.exit()

    def _args_input(self):
        self.in_stream = self.args.input
        if self._tsdata is not None:
            self.args.input = self._tsdata
        else:
            self._tsdata = self.args.input

    def _args_sidecar(self):
        Path(self.args.sidecar_file).touch()

    def _args_hls_tag(self):
        tag_map = {
            "x_scte35": self.scte35.x_scte35,
            "x_cue": self.scte35.x_cue,
            "x_daterange": self.scte35.x_daterange,
            "x_splicepoint": self.scte35.x_splicepoint,
        }
        if self.args.hls_tag not in tag_map:
            raise ValueError(f"hls tag  must be in {tag_map.keys()}")
        self.scte35.tag_method = tag_map[self.args.hls_tag]

    def _args_output_dir(self):
        if not os.path.isdir(self.args.output_dir):
            os.mkdir(self.args.output_dir)

    def _chk_flags(self, flags):
        if flags:
            return True in flags
        return None

    def _args_flags(self):
        """
        I really expected to do more here.
        """
        flags = deque([self.args.program_date_time, self.args.delete, self.args.replay])
        #       if self._chk_flags(flags):
        flags.popleft()  # pop self.args.program_date_time
        if self._chk_flags(flags):
            self.args.live = True
            self.window.delete = True
        flags.popleft()  # pop self.args.delete
        flags.popleft()  # pop self.args.replay

    def _args_window_size(self):
        """
        _args_window_size sets sliding window size
        if stream is live
        """
        if self.args.live:
            self.window.size = self.args.window_size
            if self.window.size > 10:
                self.window.size = 10  # cap window size
                blue(f"window size adjusted to {self.window.size}")

    def _args_continue_m3u8(self):
        """
        _args_continue_m3u8 continues
        a previously created index.m3u8
        works with live or vod
        """
        if self.args.continue_m3u8:
            self.continue_m3u8()

    def apply_args(self):
        """
        _apply_args  uses command line args
        to set X9K3 instance vars
        """
        self._args_version()
        self._args_input()
        self._args_sidecar()
        self._args_hls_tag()
        self._args_output_dir()
        self._args_flags()
        self._args_window_size()
        self._args_continue_m3u8()
        if isinstance(self._tsdata, str):
            self._tsdata = reader(self._tsdata)

    def _shulga_mode(self, pkt):
        """
        _shulga_mode is mpeg2 video iframe detection
        """
        if self._rai_flag(pkt):
            self._chk_splice_point()

    def m3u8uri(self):
        """
        m3u8uri return full path to the output index.m3u8
        """
        return self.mk_uri(self.args.output_dir, self.m3u8)

    @staticmethod
    def _rai_flag(pkt):
        """
        _rai_flag random access indicator flag
        """
        return pkt[5] & 0x40

    @staticmethod
    def mk_uri(head, tail):
        """
        mk_uri is used to create local filepaths
        """
        sep = "/"
        if len(head.split("\\")) > len(head.split("/")):
            sep = "\\"
        if not head.endswith(sep):
            head = head + sep
        return f"{head}{tail}"

    @staticmethod
    def clobber_file(the_file):
        """
        clobber_file  blanks the_file
        """
        with open(the_file, "w", encoding="utf8") as clobbered:
            clobbered.close()

    @staticmethod
    def _endlist(line):
        if "ENDLIST" in line:
            return True
        return False

    def _chk_splice_point(self):
        """
        _chk_splice_point checks for the slice point
        of a segment.
        """
        if self.started:
            if self.scte35.cue_time:
                self.scte35.cue_time = round(self.scte35.cue_time, 6)
                if self.started < self.scte35.cue_time <= self.now < self.next_start:
                    blue(f"self.started {self.started}")
                    blue(f"scte35.cue_time {self.scte35.cue_time}")
                    blue(f"self.now\t{self.now}")
                    red(f" DIFF: {round(self.now -self.scte35.cue_time,6)}")
                    blue(f"Scheduled self.next_start {self.next_start}")
                    self.next_start = self.scte35.cue_time = self.now
                    blue(f"Updated self.next_start   {self.next_start}")
                    blue("")
                    blue(f"scte35.cue_time {self.scte35.cue_time}")
                    blue(f"self.now\t{self.now}")
            if self.now >= self.next_start:
                self.next_start = self.now
                self._write_segment()
                self.scte35.mk_cue_state()

    def _chk_cue_time(self):
        if self.scte35.cue:
            self.scte35.cue_time = self._adjusted_pts(self.scte35.cue)

    def _chk_iframe(self, pkt):
        if self.iframer.parse(pkt):
            self.load_sidecar()
            self._chk_sidecar_cues()
            self._chk_splice_point()

    def _chk_live(self, seg_time):
        if self.args.live:
            self.window.slide_panes()
            if not self.args.no_throttle:
                self.timer.throttle(seg_time)
            self._discontinuity_seq_plus_one()

    def _chk_pdt_flag(self, segment_data):
        if self.args.program_date_time:
            iso8601 = f"{datetime.datetime.utcnow().isoformat()}Z"
            segment_data.add_tag("#EXT-X-PROGRAM-DATE-TIME", f"{iso8601}")

    def _clear_endlist(self, lines):
        return [line for line in lines if not self._endlist(line)]

    def is_byterange(self):
        """
        is byterange returns True if m3u8 is byterange.
        """
        if self.args.byterange and ".ts" in self.args.input:
            return True
        return False

    def _header(self):
        """
        header generates the m3u8 header lines
        """
        m3u = "#EXTM3U"
        m3u_version = "#EXT-X-VERSION:4"
        target = f"#EXT-X-TARGETDURATION:{int(self.args.time+1)}"
        seq = f"#EXT-X-MEDIA-SEQUENCE:{self.media_seq}"
        dseq = f"#EXT-X-DISCONTINUITY-SEQUENCE:{self.discontinuity_sequence}"
        x9k3v = f"#EXT-X-X9K3-VERSION:{version()}"
        bumper = ""
        return "\n".join(
            [
                m3u,
                m3u_version,
                target,
                seq,
                dseq,
                x9k3v,
                bumper,
            ]
        )

    def _add_discontinuity(self, segment_data):
        """
        _add_discontinuity adds a discontinuity tag.
        """
        if not self.args.no_discontinuity:
            segment_data.add_tag("#EXT-X-DISCONTINUITY", None)
        self._add_adrian(segment_data)

    def _add_adrian(self, segment_data):
        if not self.args.no_adrian_is_cool_tags_at_splice_points_because_I_suck:
            segment_data.add_tag("#EXT-X-ADRIAN-IS-COOL", None)

    def _add_cue_tag(self, segment_data):
        """
        _add_cue_tag adds SCTE-35 tags,
        auto CUE-INs, and discontinuity tags.
        """
        if self.scte35.break_timer is not None:
            if self.scte35.break_timer >= self.scte35.break_duration:
                self.scte35.break_timer = None
                self.scte35.cue_state = "IN"
        tag = self.scte35.mk_cue_tag()
        if tag:
            if self.scte35.cue_state in ["OUT", "IN"]:
                self._add_discontinuity(segment_data)
            kay = tag
            vee = None
            if ":" in tag:
                kay, vee = tag.split(":", 1)
            segment_data.add_tag(kay, vee)
            print2(f"{kay} = {vee}")

    def _mk_segment_data_tags(self, segment_data, seg_time):
        self._add_cue_tag(segment_data)
        self._chk_pdt_flag(segment_data)
        segment_data.add_tag("# Start", f" @{self.started}")
        segment_data.add_tag("#EXTINF", f"{seg_time},")
        if self.is_byterange():
            tag = "#EXT-X-BYTERANGE"
            val = f"{self.now_byte - self.started_byte}@{self.started_byte}"
            segment_data.add_tag(tag, val)

    def _mk_segment_data(self, seg_file, seg_name, seg_time):
        segment_data = SegmentData(seg_file, seg_name, self.segnum)
        if self.first_segment:
            if self.args.replay or self.args.continue_m3u8:
                self._add_discontinuity(segment_data)
        if self.skipped_segment:  # set when unable to parse a segment
            self._add_discontinuity(
                segment_data
            )  # add disco tag because pts will jump.
            self.skipped_segment = False
        self._mk_segment_data_tags(segment_data, seg_time)
        self.window.slide_panes(segment_data)

    def _print_segment_details(self, seg_name, seg_time):
        if not self.started:
            return
        one = f"{seg_name}:   start: {self.started:.3f}   "
        two = f"end: {self.now:.3f}   duration: {seg_time:.3f}"
        print2(f"{one}{two}")

    def _write_segment_file(self, seg_name):
        with open(seg_name, "wb") as seg:
            if self.pat_pkt:
                seg.write(self.pat_pkt)
                seg.write(self.pmt_pkt)
            seg.write(self.active_segment.getbuffer())

    def _write_segment(self):
        if not self.segnum:
            self.segnum = 0
        seg_file = f"seg{self.segnum}.ts"
        seg_name = self.mk_uri(self.args.output_dir, seg_file)
        if self.is_byterange():
            seg_name = self.args.input
            seg_file = self.args.input
        seg_time = round((self.now - self.started), 3)
        if seg_time <= 0:
            return
        if not self.is_byterange():
            self._write_segment_file(seg_name)
            if seg_time > self.args.time + 2:
                stuff = f"Verifying {seg_name} time of {seg_time}"
                blue(stuff)
                s = Segment(seg_name)
                s.decode()
                if s.pts_start and s.pts_last:
                    seg_time = round(s.pts_last - s.pts_start, 3)
                    stuff = f"Setting {seg_name} time to {seg_time}"
                    blue(stuff)
        self._mk_segment_data(seg_file, seg_name, seg_time)
        self._write_m3u8()
        self._print_segment_details(seg_name, seg_time)
        if self.scte35.break_timer is not None:
            self.scte35.break_timer += seg_time
        self.scte35.chk_cue_state()
        self._chk_live(seg_time)
        self._start_next_start(pts=self.now)
        self.started_byte = self.now_byte

    def _write_m3u8(self):
        self.media_seq = self.window.panes[0].num
        self._discontinuity_seq_plus_one()
        with open(self.m3u8uri(), "w+", encoding="utf8") as m3u8:
            m3u8.write(self._header())
            m3u8.write(self.window.all_panes())
            self.segnum += 1
            self.first_segment = False
        self.active_segment = io.BytesIO()
        self.window.slide_panes()

    def load_sidecar(self):
        """
        load_sidecar reads (pts, cue) pairs from
        the sidecar file and loads them into X9K3.sidecar
        """
        if self.args.sidecar_file:
            with reader(self.args.sidecar_file) as sidefile:
                sidelines = sidefile.readlines()
                if sidelines == self.last_sidelines:
                    return
                for line in sidelines:
                    line = line.decode().strip().split("#", 1)[0]
                    if line:
                        pts, data = line.split(",", 1)
                        cue = Cue(data)
                        insert_pts = self._adjusted_pts(cue)
                        if insert_pts ==-1.0:
                            insert_pts = pts
                        line = f"{insert_pts},{data}"
                        blue(f"loading  {line}")
                        if insert_pts == 0.0:
                            line = f"{self.now},{data}"
                        self.add2sidecar(line)
                sidefile.close()
                self.last_sidelines = sidelines
            self.clobber_file(self.args.sidecar_file)

    def add2sidecar(self, line):
        """
        add2sidecar add insert_pts,cue to the deque
        """
        insert_pts, cue = line.split(",", 1)
        insert_pts = float(insert_pts)
        if [insert_pts, cue] not in self.sidecar:
            self.sidecar.append([insert_pts, cue])
            self.sidecar = deque(sorted(self.sidecar, key=itemgetter(0)))

    def _chk_sidecar_cues(self):
        """
        _chk_sidecar_cues checks the insert pts time
        for the next sidecar cue and inserts the cue if needed.
        """
        if self.sidecar:
            for s in list(self.sidecar):
                splice_pts = float(s[0])
                splice_cue = s[1]
                #      if self.started:
                if self.started <= splice_pts <= self.now:
                    self.sidecar.remove(s)
                    self.scte35.cue = Cue(splice_cue)
                    self.scte35.cue.decode()
                    self.scte35.cue.show()
                    self._chk_cue_time()
                    self._chk_splice_point()

    def _discontinuity_seq_plus_one(self):
        if self.window.panes:
            if "#EXT-X-DISCONTINUITY" in self.window.panes[0].tags:
                if len(self.window.panes) >= self.window.size:
                    self.discontinuity_sequence += 1
            if "#EXT-X-DISCONTINUITY" in self.window.panes[-1].tags:
                self._reset_stream()

    def _reset_stream(self):
        self.started = None
        self.next_start = None

    def _start_next_start(self, pts=None):
        """
        _start_next_start sets when the current
        segment starts and ends.
        """
        rollover = self.ROLLOVER / 90000.0
        if pts is not None:
            self.started = pts
        else:
            self.started = self.next_start
        self.next_start = self.started + self.args.time
        if self.next_start + self.args.time > rollover:
            self._reset_stream()

    def _adjusted_pts(self, cue):
        pts = -1.0
        if "pts_time" in cue.command.get():
            pts = cue.command.pts_time
            pts_adjust = cue.info_section.pts_adjustment
            pts = (pts + pts_adjust) % self.as_90k(self.ROLLOVER)
        return round(pts, 6)

    def _parse_scte35(self, pkt, pid):
        """
        _parse_scte35 overrides the inherited method.
        """
        if self.args.exclude_mpegts:
            return None
        cue = super()._parse_scte35(pkt, pid)
        if cue:
            self._chk_cue_time()
            self.add2sidecar(f"{self._adjusted_pts(cue)}, {cue.encode()}")
        return cue

    def _parse_tables(self, pkt, pid):
        """
        _parse_tables parse for
        PAT, PMT,  and SDT tables
        based on pid of the pkt
        """
        if pid in self.pids.pmt:
            self.pmt_pkt = pkt
        if pid == 0:
            self.pat_pkt = pkt
        super()._parse_tables(pkt, pid)

    def _parse(self, pkt):
        """
        _parse is run on every packet.
        """
        super()._parse(pkt)
        self.now_byte += len(pkt)
        pkt_pid = self._parse_info(pkt)
        self.now = self.pid2pts(pkt_pid)
        if not self.started:
            self._start_next_start(pts=self.now)
        if self.started and self._pusi_flag(pkt):
            if self.args.shulga:
                self._shulga_mode(pkt)
            else:
                self._chk_iframe(pkt)
        if not self.is_byterange():
            self.active_segment.write(pkt)

    def _last_buff(self):
        """
        _last_buff writes antthing left in the
        active_segment buffer for the last segment.
        """
        buff = self.active_segment.getvalue()
        if buff:
            self._write_segment()
            time.sleep(0.3)

    def addendum(self):
        """
        addendum post stream parsing related tasks.
        """
        self._last_buff()
        if not self.args.live:
            with open(self.m3u8uri(), "a", encoding="utf8") as m3u8:
                m3u8.write("#EXT-X-ENDLIST")

    def _is_stream(self):
        """
        _is_stream determine if input is an mpegts stream.
        """
        if self.args.input == sys.stdin.buffer:
            return True
        if self.args.input.startswith(("srt://", "udp://")):
            return True
        packet = self._tsdata.read(188)
        if packet.startswith(b"#"):
            return False
        self._parse(packet)
        return True
        #return False

    def decode(self, func=False):
        """
        decode applies any set args,
        and starts parsing.
        """
        self.apply_args()
        self.timer.start()
        if self._is_stream():
            super().decode()
        else:
            self.decode_m3u8(self.args.input)
        self.addendum()

    def _retag(self, line, segment_data):
        """
        _retag when continuing an m3u8
        retag segment_data tags from line.
        """
        for tag in ["#EXT-X-SPLICEPOINT-SCTE35", "#EXT-X-SCTE35", "#EXT-X-DATERANGE"]:
            if tag in line:
                segment_data.tags[tag] = line.replace(tag + ":", "", 1)
        for tag in ["#EXT-X-CUE-IN", "#EXT-X-DISCONTINUITY"]:
            if tag in line:
                segment_data.tags[tag] = None

    def _reload_segment_data(self, segment):
        tmp_segnum = int(segment.relative_uri.split("seg")[1].split(".")[0])
        segment_data = SegmentData(
            segment.relative_uri,
            segment.media,
            tmp_segnum,
        )
        self.segnum = tmp_segnum
        for this in ["#EXT-X-X9K3-VERSION", "#EXT-X-ENDLIST"]:
            if this in segment.tags:
                segment.tags.pop(this)
        segment_data.tags = segment.tags

        if "#EXTINF" in segment.tags:
            segment_data.tags["#EXTINF"] = f'{pif(segment.tags["#EXTINF"]):.6},'
        for line in segment.lines:
            self._retag(line, segment_data)
        self.window.slide_panes(segment_data)

    def _reload_m3u8(self):
        """
        m3u8_reload is called when the continue_m3u8 option is set.
        """
        m3 = M3uFu()
        m3.window_size = None
        tmp_name = self.mk_uri(self.args.output_dir, "tmp.m3u8")
        with open(tmp_name, "w", encoding="utf8") as tmp_m3u8:
            with open(self.m3u8uri(), "r", encoding="utf8") as m3u8:
                tmp_m3u8.write("\n".join(m3u8.readlines()))
                tmp_m3u8.write("\n#EXT-X-ENDLIST\n")
        m3.m3u8 = tmp_name
        m3.decode()
        if "#EXT-X-DISCONTINUITY-SEQUENCE" in m3.headers:
            self.discontinuity_sequence = m3.headers["#EXT-X-DISCONTINUITY-SEQUENCE"]
        segments = list(m3.segments)
        m3.segments[-1].tags["#EXT-X-DISCONTINUITY"] = None
        for segment in segments:
            self._reload_segment_data(segment)
        # if self.window.panes:
        if self.args.live:
            self.window.slide_panes()
        os.unlink(tmp_name)
        self.first_segment = True

    def continue_m3u8(self):
        """
        continue_m3u8 reads self.discontinuity_sequence
        and self.segnum from an existing index.m3u8.
        """
        if os.path.isfile(self.m3u8uri()):
            self._reload_m3u8()
            self.segnum += 1
            blue(f"Continuing {self.m3u8uri()} @ segment number {self.segnum}")

    def _parse_m3u8_media(self, media):
        """
        _parse_m3u8_media parse a segment from
        a m3u8 input file if it has not been parsed.
        """
        if "master.m3u8" in media:
            return
        if media != self.first_on_page:
            try:
                self._tsdata = reader(media)
                for pkt in self.iter_pkts():
                    self._parse(pkt)
                self._tsdata.close()
            except ERR:
                blue(f"skipping {media}")
                self.skipped_segment = True

    def decode_m3u8(self, manifest=None):
        """
        decode_m3u8 is called when the input file is a m3u8 playlist.
        """
        based = manifest.rsplit("/", 1)
        if len(based) > 1:
            base_uri = f"{based[0]}/"
        else:
            base_uri = ""
        #        while True:
        media_list = deque()
        with reader(manifest) as manifesto:
            m3u8 = manifesto.readlines()
            for line in m3u8:
                if not line:
                    blue("NO LINE")
                    return False
                line = _clean_line(line)
                if self._endlist(line):
                    return False
                if line.startswith("#"):
                    media = None
                else:
                    media = line
                    if base_uri not in media:
                        media = base_uri + media
                if media:
                    media_list.append(media)
                    self._parse_m3u8_media(media)
            self.first_on_page = media_list[0]
            media_list = deque()


class Timer:
    """
    Timer class instances are used for
    segment duration, and live throttling.
    """

    def __init__(self):
        self.started = time.time()
        self.begin = None
        self.end = None
        self.lap_time = None

    def start(self, begin=None):
        """
        start starts the timer
        """
        self.begin = begin
        if not self.begin:
            self.begin = time.time()
        self.end = None
        self.lap_time = None

    def stop(self, end=None):
        """
        stop stops the timer
        """
        self.end = end
        if not self.end:
            self.end = time.time()
        self.lap_time = self.end - self.begin

    def elapsed(self, now=None):
        """
        elapsed returns the elapsed time
        """
        if not now:
            now = time.time()
        return now - self.started

    def throttle(self, seg_time, begin=None, end=None):
        """
        throttle is called to slow segment creation
        to simulate live streaming.
        """
        self.stop(end)
        diff = round((seg_time - self.lap_time) * 0.99, 2)
        if diff > 0:
            blue(f"throttling {diff}")
            time.sleep(diff)
        self.start(begin)


class SegmentData:
    """
    Class to hold hls segment tags
    for a segment.
    """

    def __init__(self, file, name, num):
        self.tags = {}
        self.file = file
        self.name = name
        self.num = num

    def __repr__(self):
        return str(vars(self))

    def get(self):
        """
        get returns the segment_data data formated.
        """
        this = []
        for kay, vee in self.tags.items():
            if vee is None:
                this.append(kay)
            else:
                this.append(f"{kay}:{vee}")
        this.append(self.file)
        this.append("")
        this = "\n".join(this)
        return this

    def add_tag(self, quay, val):
        """
        add_tag appends key and value for a hls tag
        """
        self.tags[quay] = val


def _clean_line(line):
    if isinstance(line, bytes):
        line = line.decode(errors="ignore")
    line = line.replace("\n", "").replace("\r", "")
    return line


def decode_playlist(playlist):
    """
    decode_playlist parses a playlist file
    and segments all the media into 1 stream.
    A playlist file is a list of media OR media,sidecar lines
    Example:

    /home/a/video.ts
    /home/a/othervideo.ts,/home/a/other_sidecar.txt
    https://futzu.com/xaa.ts

    """
    comma = ","
    octothorpe = "#"
    first = True
    sidecar = None
    with reader(playlist) as plist:
        for line in plist.readlines():
            if not line:
                break
            line = _clean_line(line)
            media = line.split(octothorpe)[0]
            if media:
                if comma in media:
                    media, sidecar = media.split(comma)
                blue(f"loading media {media}")
            x9 = X9K3()
            if sidecar:
                Path(sidecar).touch()
                blue(f"loading sidecar file {sidecar}")
                x9.args.sidecar_file = sidecar
            x9.args.input = media
            if first:
                x9.decode()
                first = False
            else:
                x9.continue_m3u8()
                x9.decode()


def cli():
    """
    cli provides one function call for running X9K3.
    """
    args = argue()
    if isinstance(args.input, str) and (".playlist" in args.input):
        decode_playlist(args.input)
    else:
        x9 = X9K3()
        x9.decode()
        while args.replay:
            x9 = X9K3()
            x9.continue_m3u8()
            x9.decode()


if __name__ == "__main__":
    cli()



"""
abr.py
"""

from pathlib import Path
import os
import sys
import time
import multiprocessing as mp
from threefive import reader
from .x9k3 import X9K3, argue

mp.set_start_method("fork")


# import cProfile


class ABR:
    """
    The ABR class starts a X9MP process for each rendition
    and manages them.
    """

    def __init__(self, m3u8_list):
        self.master = None
        self.m3u8_list = m3u8_list # list of m3u8 manifests for renditions [index.m3u8, index2.m3u8]
        self.args = argue()
        self.sidecar = self.args.sidecar_file
        self.side_files = []
        self.last_stat = 0
        self.procs = []

    def add_rendition(self, m3u8, dir_name, rendition_sidecar=None):
        """
        add_rendition starts a process for each rendition and
        creates a pipe for each rendition to receive SCTE-35.
        """
        p = mp.Process(
            target=mp_run,
            args=(m3u8, dir_name, rendition_sidecar),
        )
        p.start()
        print(f"Rendition Process Started {dir_name}")
        self.procs.append(p)

    def load_sidecar(self):
        """
        load_sidecar reads (pts, cue) pairs from
        the sidecar file and loads them into X9K3.sidecar
        if live, blank out the sidecar file after cues are loaded.
        """
        with reader(self.sidecar) as sidefile:
            these_lines = sidefile.readlines()
        with open(self.sidecar, "w", encoding="utf8") as sided:
            sided.close()
        if these_lines:
            for side_file in self.side_files:
                print(f"Updating {side_file}")
                with open(side_file, "wb") as side:
                    side.writelines(these_lines)

    def _mk_rendition_sidecar(self, dir_name):
        rendition_sidecar = None
        if self.args.sidecar_file:
            _, tail = os.path.split(self.sidecar)
            rendition_sidecar = dir_name + "/" + tail
            Path(rendition_sidecar).touch()
            self.side_files.append(rendition_sidecar)
        return rendition_sidecar

    @staticmethod
    def clobber_file(the_file):
        """
        clobber_file  blanks the_file
        """
        with open(the_file, "w", encoding="utf8") as clobbered:
            clobbered.close()

    def _chk_master_sidecar(self):
        Path(self.sidecar).touch()
        side_stat = os.stat(self.sidecar).st_mtime
        if side_stat != self.last_stat:
            self.load_sidecar()
            self.clobber_file(self.sidecar)
            side_stat = os.stat(self.sidecar).st_mtime
            self.last_stat = side_stat

    def go(self):
        """
        go writes the new master.m3u8.
        """
        dir_name = 0
        with open(
            self.args.output_dir + "/master.m3u8", "w", encoding="utf-8"
        ) as master:
            master.write("#EXTM3U\n#EXT-X-VERSION:6\n\n")
            for m3u8 in self.m3u8_list:
                if "#EXT-X-STREAM-INF" in m3u8.tags:
                    master.write("\n".join(m3u8.lines[:-1]))
                    master.write("\n")
                    dn = self.args.output_dir + "/" + str(dir_name)
                    if not os.path.isdir(dn):
                        os.mkdir(dn)
                    master.write(f"{dir_name}/index.m3u8\n")
                    rendition_sidecar = self._mk_rendition_sidecar(dn)
                    self.add_rendition(m3u8, dn, rendition_sidecar)
                    dir_name += 1
                else:  # Copy over stuff like "#EXT-X-MEDIA-TYPE"
                    if len(m3u8.lines) > 1:
                        master.write("\n".join(m3u8.lines[:-1]))
                        media_uri = m3u8.lines[-1]
                        master.write(f"{media_uri}\n")
                        master.write("\n")
                    else:
                        master.write(m3u8.lines[0])
                        master.write("\n")
        while True in [p.is_alive() for p in self.procs]:
            self._chk_master_sidecar()
            time.sleep(0.2)
        sys.exit()


def mk_x9mp(manifest, dir_name, rendition_sidecar):
    """
    mk_x9mp generates an X9MP instance and
    sets default values
    """
    x9mp = X9K3()
    x9mp.args = argue()
    x9mp.args.output_dir = dir_name
    x9mp.args.input = manifest.media
    x9mp.args.sidecar_file = rendition_sidecar
    return x9mp


def mp_run(manifest, dir_name, rendition_sidecar=None):
    """
    mp_run is the process started for each rendition.
    """
    args = argue()
    x9mp = mk_x9mp(manifest, dir_name, rendition_sidecar)
    x9mp.decode()
    while args.replay:
        x9mp = mk_x9mp(manifest, dir_name, rendition_sidecar)
        x9mp.args.continue_m3u8 = True
        x9mp.continue_m3u8()
        x9mp.decode()
    return False



#!/usr/bin/env python3

"""
x9k3/x9mp.py

cli tool for ABR
"""

import os
import sys
from m3ufu import M3uFu
from threefive import reader, blue
from .abr import ABR
from .x9k3 import argue, version
from .x9k3 import cli as x9cli


def do(args):
    """
    do runs x9k3.ABR programmatically.

    """
    fu = M3uFu()
    if not args.input:
        print("input source required (Set args.input)")
        sys.exit()
    fu.m3u8 = args.input
    if not os.path.isdir(args.output_dir):
        os.mkdir(args.output_dir)
    fu.decode()
    abr = ABR(fu.segments)
    abr.go()


def not_master(m3u8):
    """
    not_master returns True if
    m3u8 is not a master.m3u8
    """
    if m3u8 == sys.stdin.buffer:
        return True
    if m3u8.startswith(("srt://", "udp://")):
        return True


def is_master(m3u8):
    """
    is_master master.m3u8 detection
    """
    if not_master(m3u8):
        return False
    return b"#EXT-X-STREAM-INF" in reader(m3u8).read(1316)


def cli():
    """
    cli provides one function call
    to handle x9k3 with ABR support

    from x9k3.x9mp import cli
     cli()

    """
    args = argue()
    if args.version:
        print(version())
        sys.exit()
    _ = {blue(f"{k} = {v}") for k, v in vars(args).items()}
    if is_master(args.input):
        do(args)
    else:
        x9cli()


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3

import setuptools
import x9k3

with open("README.md", "r") as fh:
    readme = fh.read()

setuptools.setup(
    name="x9k3",
    version=x9k3.version(),
    author="Adrian of Doom and a Cast of Thousands",
    author_email="spam@iodisco.com",
    description="ABR HLS Segmenter with SCTE-35 Injecton",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/superkabuki/x9k3",
    packages=setuptools.find_packages(),
    scripts=["bin/x9k3","bin/adbreak3","bin/adinterval3",],
    platforms="all",
    install_requires=[
        "threefive >= 3.0.71",
        "m3ufu >= 0.0.97",
    ],
    classifiers=[
        "License :: OSI Approved :: Sleepycat License",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "Operating System :: POSIX :: BSD :: OpenBSD",
        "Operating System :: POSIX :: BSD :: NetBSD",
        "Operating System :: POSIX :: Linux",
        "Topic :: Multimedia :: Video",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    python_requires=">=3.9",
)
